@@ -1,12 +1,13 @@
 # SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
 # SPDX-License-Identifier: MIT
+import io
 import shutil
 import sys
 from pathlib import Path
 from tempfile import TemporaryDirectory
-from subprocess import check_call, run as sprun, CalledProcessError, PIPE, \
-    STDOUT
-from typing import Optional, List, Union, Iterator
+from subprocess import run as run_process, CalledProcessError, PIPE, STDOUT, \
+    CompletedProcess
+from typing import Optional, List, Union, Iterator, Type
 import venv
 import os
 
@@ -21,7 +22,10 @@ def print_command(cmd, title, at):
           ('=' * (80 - len(at) - len(title) - 4 - 4)) +
           ' ' + at + ' ==')
 
-    print(' '.join(repr(arg) for arg in cmd))
+    if isinstance(cmd, str):
+        print(cmd)
+    else:
+        print(' '.join(repr(arg) for arg in cmd))
     print('=' * 80)
     print()
 
@@ -91,24 +95,39 @@ def find_latest_wheel(parent_dir: Path) -> Path:
 
 
 class Runner:
-    """Runs commands with the same executable file. Prints the commands
-    to the stdout with the same 'at' comments."""
+    """Runs commands with the same executable file (until reformat_args=False).
+    Prints the commands to the stdout with the same 'at' comments."""
 
     def __init__(self, exe, at):
         self.exe = exe
         self.at = at
 
-    def run(self, args: Union[str, List[str]], title=None, cwd=None,
-            exception=None):
-        args_list = args.split() if isinstance(args, str) else args
-        args_list = [self.exe] + args_list
+    def run(self,
+            args: Union[str, List[str]],
+            title: str,
+            reformat_args: bool = True,
+            cwd: Union[Path, str] = None,
+            exception: Type[BaseException] = None,
+            executable: str = None,
+            shell: bool = False):
+
+        if reformat_args:
+            args_list = args.split() if isinstance(args, str) else args
+            args_list = [self.exe] + args_list
+        else:
+            args_list = args
+
         print_command(cmd=args_list, at=self.at, title=title)
 
-        cp = sprun(args_list, cwd=cwd, encoding=sys.stdout.encoding,
-                   stdout=PIPE, stderr=STDOUT,
-                   universal_newlines=True)
+        cp = run_process(args_list, cwd=cwd, encoding=sys.stdout.encoding,
+                         stdout=PIPE, stderr=STDOUT,
+                         executable=executable,
+                         shell=shell,
+                         universal_newlines=True)
 
-        print(cp.stdout.strip())
+        output = cp.stdout.rstrip()
+        if output:
+            print(output)
 
         if cp.returncode != 0:
             cpe = CalledProcessError(cp.returncode, args_list, cp.stdout,
@@ -164,6 +183,16 @@ class BuildCleaner:
         self.cleanup()
 
 
+def run_as_bash_script(script: str, input: bytes = None) -> CompletedProcess:
+    """Runs the provided string as a .sh script."""
+    # almost same as in VIEN
+
+    # we need executable='/bin/bash' for Ubuntu 18.04, it will run
+    # '/bin/sh' otherwise. For MacOS 10.13 it seems to be optional
+    return run_process(script, shell=True, executable='/bin/bash',
+                       input=input)
+
+
 class Package:
     """During initialization, this object creates a .whl distribution
     and installs the packages from the distribution into a test virtual
@@ -177,6 +206,7 @@ class Package:
         self._close_us = list()
         self._installer: Optional[Runner] = None
         self.project_source_dir = Path(project_dir).absolute()
+        self.installer_venv: Optional[TempVenv] = None
 
     def __enter__(self):
         self.init()
@@ -189,14 +219,18 @@ class Package:
 
         builder = Runner(builder_python, at='builder venv')
 
+        # INSTALLING BUILD ####################################################
+
         builder.run('-m pip install --upgrade pip',
                     title='Upgrading pip',
                     exception=CannotInitializeEnvironment)
-        builder.run('-m pip install --upgrade build twine',
-                    title='Upgrading build and twine',
+        builder.run('-m pip install --upgrade build',
+                    title='Installing build',
                     exception=CannotInitializeEnvironment)
 
         with TemporaryDirectory() as temp_dist_dir:
+            # BUILDING ########################################################
+
             with BuildCleaner(self.project_source_dir):
                 builder.run(
                     ['-m', 'build', '--outdir', temp_dist_dir, '--wheel'],
@@ -208,15 +242,24 @@ class Package:
             whl = whl.absolute()
             print(f'Latest wheel: {whl}')
 
+            # TWINE CHECK #####################################################
+
+            builder.run('-m pip install --upgrade twine',
+                        title='Installing twine',
+                        exception=CannotInitializeEnvironment)
+
             # running twine checks on the new file
             builder.run(['-m', 'twine', 'check',
                          os.path.join(temp_dist_dir, '*'), '--strict'],
                         title='Twine check',
                         exception=TwineCheckFailed)
 
-            installer_venv = TempVenv()
-            self._close_us.append(installer_venv)
-            installer_python = installer_venv.__enter__()
+            # TEST VENV #######################################################
+
+            self.installer_venv = TempVenv()
+
+            self._close_us.append(self.installer_venv)
+            installer_python = self.installer_venv.__enter__()
             self._installer = Runner(installer_python, at='installer venv')
 
             self._installer.run('-m pip install --upgrade pip',
@@ -235,15 +278,82 @@ class Package:
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.cleanup(exc_type, exc_val, exc_tb)
 
+    @staticmethod
+    def _output(cp: CompletedProcess, rstrip: bool = True) -> str:
+        o: str = cp.stdout
+        if rstrip:
+            o = o.rstrip()
+        return o
+
     def run_python_code(self, code: str, rstrip: bool = True):
         with TemporaryDirectory() as temp_current_dir:
             cp = self._installer.run(['-c', code],
-                                     title="Running code (cwd is temp dir)",
+                                     title="Running Python code (cwd is temp dir)",
                                      cwd=temp_current_dir,
                                      exception=CodeExecutionFailed)
 
-            output = cp.stdout
-            if rstrip:
-                output = output.rstrip()
+            return self._output(cp, rstrip)
 
-            return output
+    @property
+    def _can_run_bash(self):
+        return os.path.exists("/bin/bash")
+
+    def _run_bash_code(self, code: str, rstrip: bool = True):
+        with TemporaryDirectory() as temp_current_dir:
+            activate = os.path.join(self.installer_venv.venv_dir, 'bin',
+                                    'activate')
+
+            code = '\n'.join(["#!/bin/bash",
+                              "set -e",
+                              f'source "{activate}"',
+                              code])
+
+            # we need executable='/bin/bash' for Ubuntu 18.04, it will run
+            # '/bin/sh' otherwise. For MacOS 10.13 it seems to be optional
+            cp = self._installer.run(
+                code,
+                reformat_args=False,
+                title="Running Bash code (cwd is temp dir)",
+                cwd=temp_current_dir,
+                shell=True, executable='/bin/bash',
+                # input=code,
+                exception=CodeExecutionFailed)
+
+            return self._output(cp, rstrip)
+
+    def _run_cmdexe_code(self, code: str, rstrip: bool = True):
+        """Runs command in cmd.exe"""
+        with TemporaryDirectory() as temp_current_dir:
+            # file that activates the venv
+            activate_bat = os.path.join(
+                self.installer_venv.venv_dir,
+                'Scripts',
+                'activate.bat')
+            assert os.path.exists(activate_bat), "activate.bat not found"
+
+            # temp file with commands to run
+            temp_bat_file = Path(temp_current_dir) / "_run_cmdexe_code.bat"
+            temp_bat_file.write_text(
+                '\n'.join([f"CALL {activate_bat}",
+                           code]))
+
+            # todo param /u formats output as unicode?
+            cp = self._installer.run(
+                ["cmd.exe", "/q", "/c", str(temp_bat_file)],
+                reformat_args=False,
+                title="Running code in cmd.exe (cwd is temp dir)",
+                cwd=temp_current_dir,
+                shell=True,  # executable='/bin/bash',
+                # input=code,
+                exception=CodeExecutionFailed)
+
+            return self._output(cp, rstrip)
+
+    def run_shell_code(self, code: str, rstrip: bool = True) -> str:
+
+        if os.name == 'nt':
+            method = self._run_cmdexe_code
+        else:
+            method = self._run_bash_code
+
+        return method(code, rstrip=rstrip)
